import numpy as np
from umap import UMAP
from sklearn.decomposition import PCA, TruncatedSVD, IncrementalPCA
from scipy.sparse import issparse
from sklearn.base import BaseEstimator
from umap.umap_ import optimize_layout_euclidean
from scipy.sparse import coo_matrix

from song.util import find_spread_tightness, \
    bulk_grow_with_drifters, embed_batch_epochs, \
    train_for_batch_batch, sq_eucl_opt, get_closest_for_inputs, get_fast_knn

INT32_MIN = np.iinfo(np.int32).min + 1
INT32_MAX = np.iinfo(np.int32).max - 1


class SONG(BaseEstimator):

    def __init__(self, n_components=2, n_neighbors=1,
                 lr=1., gamma=None, so_steps = None, mutable_graph = True,
                 spread_factor=0.9,
                 spread=1., min_dist=0.1, ns_rate=5,
                 agility=1., verbose=0,
                 max_age=3,
                 random_seed=1, epsilon=.9, a=None, b=None, final_vector_count=None, dispersion_method = 'UMAP',
                 fvc_growth=0.5, chunk_size = 1000, pc_components = 50, non_so_rate = 0, low_memory = False, sampled_batches = False, um_min_dist = 0.001, um_lr = 0.01, um_epochs = 11):

        ''' Initialize a SONG to reduce data.

        :param n_components : Dimensionality of the required output. Default set at 2 dimensions

        :param n_neighbors : Neighborhood size of the Neural (Coding Vector) Graph. Keep it low (1 or 2) for clear
         cluster separation. Higher values result in highly connected graphs, providing poor cluster separation. High
         values are more tolerant to noise.

        :param lr : Learning Rate starting value. Set at 1.0. For large data sets, smaller values may be required
        to prevent cluster drifts

        :param gamma : This variable governs the sampling schedule of the algorithm. Higher gamma will force SONG to
         start with smaller sample numbers and progressively reach the full dataset in each self-organizing iteration.
         gamma = 0. will sample the full dataset at each self-organizing iteration.

        :param so_steps : How many self organization steps are required. Higher numbers of so_steps will cause better
        visualization at the cost of running time.

        :param spread_factor : Spread Factor as defined in the work by Alahakoon et al (2000). Values in (0, 1). Higher
        values result in higher resolutions, and finer grain clustering, but inhibits time consumption performance.

        :param verbose : Print out learning process on command line.

        :param agility : A more agile map will adopt the map to new data (growth and all) faster than an inert map.
        Setting agility == 1. will treat all new data as equally weighed to old data. agility = (0, 1].

        :param spread : Hyperparameter to govern how spread-out the visualization needs to be.

        :param min_dist : Hyperparameter to govern how close together neighbors should be.

        :param epsilon : The edge-strength decay constant. Set at a small value for sparse graph generation. For dense
        graphs, use a higher value (close to 1.0 but less than 1.0).

        :param max_age : parameter to define the lowest permitted edge-strength in the graph. Higher the max_age, the
        smaller the minimum edge-strength.

        :param a : constant term for the rational quadratic kernel

        :param b : constant term for the rational quadratic kernel

        :param final_vector_count : the number of expected final coding vectors.


        :param fvc_growth: In incremental scenarios, what is the expected growth ratio of coding vectors in subsequent
        visualizations.

        '''

        self.max_epochs_per_sample = np.float32(1)
        self.ns_rate = ns_rate
        self.dim = n_components
        self.lrst = lr
        self.map_ratio = 1.5
        self.sf = spread_factor
        self.spread = spread
        self.min_dist = min_dist
        self.n_neighbors = n_neighbors
        self.random_seed = random_seed
        self.random_state = np.random.RandomState(seed=random_seed)
        self.rng_state = self.random_state.randint(INT32_MIN, INT32_MAX, 3).astype(np.int64)
        self.agility = agility
        self.gamma = gamma
        self.verbose = verbose
        self.epsilon = epsilon
        self.alpha = a
        self.beta = b
        self.fvc = final_vector_count
        self.max_age = max_age
        self.reduced_lr = 1.
        self.prototypes = None
        self.min_strength = epsilon ** ((self.dim + self.max_age))
        self.ss = so_steps
        self.non_so_rate = non_so_rate + 1
        self.prot_inc_portion = fvc_growth
        self.low_memory = low_memory
        self.trained = False
        self.dispersion_method = dispersion_method
        self.um_lr = um_lr
        self.um_epochs = um_epochs
        self.um_min_dist = um_min_dist
        self.chunk_size = chunk_size
        self.pca = None
        self.pc_components = pc_components
        self.mutable_graph = mutable_graph

        if not sampled_batches is None:
            self.sampled_batches = sampled_batches
        else:
            self.sampled_batches = low_memory

    def __repr__(self):
        return f'SONG model: n(coding vectors) = {self.W.shape[0]}, n(dimensions) = {self.W.shape[1]}, n(components) = {self.Y.shape[1]}' if self.trained else f'untrained SONG model'

    def _train_pca(self, X):

        self.pca = PCA(n_components=np.min([self.pc_components, X.shape[0], X.shape[1]]), random_state = self.random_seed)
        self.pca.fit(X.toarray() if issparse(X) else X)

    def _get_XPCA(self, X):
        nc = np.min([self.pc_components, X.shape[0], X.shape[1]])
        X_pca = np.zeros((X.shape[0], nc))
        chunk_size = 5000
        if issparse(X):
            n_chunks = X.shape[0] // chunk_size + 1

            for i in range(n_chunks):
                st = i * chunk_size
                et = (i + 1) * chunk_size

                X_c = X[st:et].toarray()
                if X_c.shape[0]:
                    X_pca[st:et] = self.pca.transform(X_c)

        else:

            X_pca = self.pca.transform(X)

        return X_pca

    def fit(self, X, reduction = 'PCA', corrected_PC = np.array([])):
        '''
        :param X: The input dataset normalized over the dataset to range [0, 1].
        :param L: If needed, provide labels for the intermediate visualizations. Must be same length as input array and
        same order
        :return: Y : The Mapped Coordinates in the desired output space (default = 2 ).
        '''
        X = X.astype(np.float32)

        if reduction == 'PCA' and X.shape[1] > self.pc_components:
            if not corrected_PC.shape[0]:
                if self.verbose:
                    print('Fitting Reduction for Neighborhood Function Calculation')

                if self.pca == None:
                    self._train_pca(X[self.random_state.permutation(X.shape[0])[:10000]])

                if self.verbose:
                    print ('reduction model fitted!')
            if not corrected_PC.shape[0]:

                X_PCA = self._get_XPCA(X)
            else:
                X_PCA = corrected_PC
        else:
            X_PCA = X
        verbose = self.verbose
        sparse = issparse(X)
        min_dist = self.min_dist
        spread = self.spread

        min_batch = 1000 if not self.sampled_batches else 10000

        if self.ss is None:
            if X.shape[0] < 100000 or not self.sampled_batches:
                self.ss = 100#(20 if X.shape[0] > 100000 else 20)
            else:
                self.ss = X.shape[0] // min_batch + 10
        self.max_its = self.ss * self.non_so_rate

        if self.alpha is None and self.beta is None:
            alpha, beta = find_spread_tightness(spread, min_dist)
            self.alpha = alpha
            self.beta = beta
        else:
            alpha = self.alpha
            beta = self.beta


        if self.sf is None:
            self.sf = np.log(4) / (2 * self.ss)

        error_scale = np.median(np.linalg.norm(X-X.mean(axis=0) if not (reduction == 'PCA') else X_PCA-X_PCA.mean(axis=0), axis=1))**2
        thresh_g = -(X.shape[1]) if not (reduction=='PCA') else -(X_PCA.shape[1]) * np.log(self.sf) * error_scale

        so_lr_st = self.lrst
        if self.prototypes is None:
            if not self.fvc is None:
                self.prototypes = self.fvc
            else:
                self.prototypes = int(np.exp(np.log(X.shape[0]) / 1.5))
        ''' Initialize coding vector weights and output coordinate system  '''
        ''' index of the last nearest neighbor : depends on output dimensionality. Higher the output dimensionality, 
                       higher the connectedness '''

        im_neix = self.dim + self.n_neighbors

        if self.trained:
            ''' When reusing a trained model, this block will be executed.'''
            W = self.W
            G = self.G
            Y = self.Y
            E_q = self.E_q
            self.prototypes += self.prototypes * self.prot_inc_portion
            self.prototypes = np.int(self.prototypes)
            if verbose:
                print('Using Trained Map')
        else:
            ''' If the model is not already initialized ...'''
            init_size = im_neix
            W = X[self.random_state.choice(X.shape[0], init_size)] + self.random_state.random_sample(
                (init_size, X.shape[1])).astype(np.float32) * 0.0001
            Y = self.random_state.random_sample((init_size, self.dim)).astype(np.float32)
            if verbose:
                print('random map initialization')
                print('Stopping at {} prototypes'.format(self.prototypes))

            ''' Initialize Graph Adjacency and Weight Matrices '''
            G = np.identity(W.shape[0]).astype(np.float32)
            E_q = np.zeros(W.shape[0]).astype(np.float32)

        lrst = self.lrst

        presented_len = X.shape[0]
        soed = 0
        lrdec = 1.
        soeds = np.arange(self.ss)
        sratios = ((soeds) * 1. / (self.ss - 1))

        batch_sizes = (X.shape[0] - min_batch) * ((sratios * 0) if self.low_memory else sratios ** (100) ) + min_batch
        epsilon = self.epsilon
        lr_sigma = np.float32(5)
        drifters = np.array([])
        order = self.random_state.permutation(X.shape[0])
        for i in range(self.max_its):
            order = self.random_state.permutation(X.shape[0]) if not self.sampled_batches else order
            if not i % self.non_so_rate:
                presented_len = int(batch_sizes[soed]) if not self.sampled_batches else min_batch
                soed += 1
            X_presented = X[order[:presented_len]] if not self.sampled_batches else X[(presented_len * i)%X.shape[0] : min((presented_len * i)%X.shape[0]+presented_len, X.shape[0])].astype(np.float32)
            X_presented_pc = X_PCA[order[:presented_len]] if not self.sampled_batches else X_PCA[(presented_len * i)%X.shape[0] : min((presented_len * i)%X.shape[0]+presented_len, X.shape[0])].astype(np.float32)

            if not i % self.non_so_rate:
                non_growing_iter = 0
                shp = np.arange(G.shape[0]).astype(np.int32)

                if self.mutable_graph and self.prototypes >= G.shape[0] and i > 0:
                    ''' Growing of new coding vectors and low-dimensional vectors '''
                    W, G, Y, E_q = bulk_grow_with_drifters(shp, E_q, thresh_g, drifters, G, W, Y)

                '''shp is an index set used for optimizing search operations'''
                shp = np.arange(G.shape[0], dtype=np.int32)

                chunk_size =  self.chunk_size if i else 100
                n_chunks = X_presented.shape[0]//chunk_size + 1
                if reduction == 'PCA' and X.shape[1] > 100:
                    W_ = self.pca.transform(W).astype(np.float32)
                else:
                    W_ = W

                for chunk in range(n_chunks):
                    chunk_st = chunk * chunk_size
                    chunk_en = chunk_st + chunk_size
                    X_chunk = X_presented[chunk_st:chunk_en].toarray() if sparse else X_presented[chunk_st:chunk_en]
                    if not X_chunk.shape[0]:
                        continue
                    if reduction == 'PCA':
                        X_chunk_ = (X_presented_pc[chunk_st:chunk_en]).astype(np.float32)
                    else:
                        X_chunk_ = X_chunk

                    pdists = sq_eucl_opt(X_chunk_, W_).astype(np.float32)
                    if verbose:
                        print(f'\r  |G| = {G.shape[0]}, |X| = {X_presented.shape[0]}, epoch = {i+1}/{self.max_its}, Training chunk {chunk + 1} of {n_chunks}', end='')
                    W, Y, G, E_q = train_for_batch_batch(X_chunk, pdists, i, self.max_its, lrst, lrdec, im_neix,
                                                         W,
                                                         self.max_epochs_per_sample, G, epsilon,
                                                         self.min_strength,
                                                         shp, Y,
                                                         self.ns_rate, alpha, beta, self.rng_state, E_q,
                                                         lr_sigma, self.reduced_lr)

                drifters = np.where(G.sum(axis=1) == 0)[0]

            else:
                if verbose:
                    print('\r Training sans SO epoch {} / {} , |X| = {} , |G| = {}  '.format(i + 1, self.max_its, X_presented.shape[0], G.shape[0]), end='')

                repeats = 1 if not soed == self.ss else 1
                for repeat in range(repeats):
                    Y = embed_batch_epochs(Y, G, self.max_its, i, alpha, beta, self.rng_state, self.reduced_lr)
                non_growing_iter += 1



        self.reduced_lr = self.reduced_lr * self.agility
        self.W = W
        self.Y = Y
        self.G = G
        self.E_q = E_q
        self.so_lr_st = so_lr_st
        self.trained = True
        if verbose:
            print('\n Done ...')
        return self

    def batch_train(self, X):
        pass

    def fit_transform(self, X, reduction = 'PCA', corrected_PC = np.array([])):
        self.fit(X, reduction, corrected_PC)
        return self.transform(X, reduction)

    def transform(self, X, reduction = 'PCA', corrected_PC = np.array([])):

        '''Adding a PCA-reduction to speed up the transform process'''

        min_dist_args, _, X_pc, W_pc = self.get_representatives(X, reduction = reduction, corrected_PC = corrected_PC)

        output = self.Y[min_dist_args]

        Y = output
        if self.dispersion_method == 'UMAP':
            Y += self.random_state.random_sample(Y.shape) * 0.001
            output = self.get_umap_dispersion(Y, X_pc.astype(np.float32), W_pc.astype(np.float32))

        return output

    def get_umap_dispersion(self, Y_init, X_pc, W_pc):
        if self.verbose:
            print('constructing graph')
        # values, rows, cols = get_fast_knn(X_pc,  W_pc, 15, self.G)
        #
        # epochs_per_sample = np.exp(-values) * 10
        # if self.verbose:
        #     print('optimizing layout')
        # Y = optimize_layout_euclidean(Y_init, Y_init.copy(), rows, cols, 20, X_pc.shape[0], epochs_per_sample, self.alpha, self.beta, self.rng_state, initial_alpha=0.01)
        self.Y_loc = Y_init.min(axis=0)
        self.Y_scale = Y_init.max(axis=0) - self.Y_loc

        Y = UMAP(init=Y_init, min_dist=self.min_dist, n_components=self.dim, spread=self.spread, learning_rate=0.01,
                 n_epochs=11).fit_transform(X_pc)
        if self.verbose:
            print('transformation done...')
        return (Y * self.Y_scale / 10.) + self.Y_loc


    def get_representatives(self, X, reduction = 'PCA', corrected_PC = np.array([])):

        if reduction == 'PCA' and X.shape[1] > self.pc_components:

            W_pc = self._get_XPCA(self.W).astype(np.float32)
            if not corrected_PC.shape[0]:
                X_pc = self._get_XPCA(X)
            else:
                X_pc = corrected_PC
            if len(X_pc.shape) == 1:
                X_pc = np.array([X_pc])

            min_dist_args = []
            min_dist_vals = []
            chunk = 1000
            n_chunks = X.shape[0] // chunk + 1
            for i in range(n_chunks):
                if self.verbose:
                    print(f'\rgetting PC of chunk {i+1}/{n_chunks}', end='')
                X_b = X_pc[i * chunk: (i + 1) * chunk]

                min_dist_pos, min_dist_val = get_closest_for_inputs(np.float32(X_b), W_pc)

                min_dist_args.extend(list(min_dist_pos))
                min_dist_vals.extend(list(min_dist_val))
            if self.verbose:
                print('')
        else:
            X_pc = X
            W_pc = self.W
            if len(X.shape) == 1:
                X = np.array([X])
            min_dist_args = []
            min_dist_vals = []

            chunk = 1000
            n_chunks = X.shape[0] // chunk + 1
            for i in range(n_chunks):
                if self.verbose:
                    print(f'\rgetting PC of chunk {i+1}/{n_chunks}', end='')
                X_b = X[i * chunk: (i + 1) * chunk].toarray() if issparse(X) else X[i * chunk: (i + 1) * chunk]
                min_dist_pos, min_dist_val = get_closest_for_inputs(np.float32(X_b), self.W)
                min_dist_args.extend(list(min_dist_pos))
                min_dist_vals.extend(list(min_dist_val))
            if self.verbose:
                print('')
        return min_dist_args, min_dist_vals, X_pc, W_pc



import numba
import numpy as np
from scipy.optimize import curve_fit
from pynndescent import NNDescent


@numba.njit("i4(i8[:])")
def fast_random_integer(random_state):
    """
    XORShift Pseudorandom Number Generator (inspired by use in UMAP)
    0xffffffff ensures 32bit truncation for faster operations

    """
    random_state[0] &= 6489292939
    random_state[0] ^= random_state[0]<<13
    random_state[0] ^= random_state[0]>>7
    random_state[0] ^= random_state[0]<<17
    random_state[0] &=  0xffffffff

    random_state[0] &= 6498787687
    random_state[1] ^= random_state[1] << 15
    random_state[1] ^= random_state[1] >> 9
    random_state[1] ^= random_state[1] << 17
    random_state[1] &= 0xffffffff

    random_state[0] &= 64654892939
    random_state[2] ^= random_state[2] << 17
    random_state[2] ^= random_state[2] >> 5
    random_state[2] ^= random_state[2] << 3
    random_state[2] &= 0xffffffff

    return random_state[0] ^ random_state[1] ^ random_state[2]

@numba.njit('f4[:,:](f4[:,:], f4[:,:])', fastmath=True, parallel=True)
def sq_eucl_opt(A, B):
    ''' function adapted from https://github.com/numba/numba-scipy/issues/38#issuecomment-623569703 '''
    C = np.empty((A.shape[0], B.shape[0]), np.float32)
    I_BLK = 32
    J_BLK = 32

    init_val_arr = np.zeros(1, A.dtype)
    init_val = init_val_arr[0]

    for ii in numba.prange(A.shape[0] // I_BLK):
        for jj in range(B.shape[0] // J_BLK):
            for i in range(I_BLK // 4):
                for j in range(J_BLK // 2):
                    acc_0 = init_val
                    acc_1 = init_val
                    acc_2 = init_val
                    acc_3 = init_val
                    acc_4 = init_val
                    acc_5 = init_val
                    acc_6 = init_val
                    acc_7 = init_val
                    for k in range(A.shape[1]):
                        acc_0 += (A[ii * I_BLK + i * 4 + 0, k] - B[jj * J_BLK + j * 2 + 0, k]) ** 2
                        acc_1 += (A[ii * I_BLK + i * 4 + 0, k] - B[jj * J_BLK + j * 2 + 1, k]) ** 2
                        acc_2 += (A[ii * I_BLK + i * 4 + 1, k] - B[jj * J_BLK + j * 2 + 0, k]) ** 2
                        acc_3 += (A[ii * I_BLK + i * 4 + 1, k] - B[jj * J_BLK + j * 2 + 1, k]) ** 2
                        acc_4 += (A[ii * I_BLK + i * 4 + 2, k] - B[jj * J_BLK + j * 2 + 0, k]) ** 2
                        acc_5 += (A[ii * I_BLK + i * 4 + 2, k] - B[jj * J_BLK + j * 2 + 1, k]) ** 2
                        acc_6 += (A[ii * I_BLK + i * 4 + 3, k] - B[jj * J_BLK + j * 2 + 0, k]) ** 2
                        acc_7 += (A[ii * I_BLK + i * 4 + 3, k] - B[jj * J_BLK + j * 2 + 1, k]) ** 2
                    C[ii * I_BLK + i * 4 + 0, jj * J_BLK + j * 2 + 0] = (acc_0)
                    C[ii * I_BLK + i * 4 + 0, jj * J_BLK + j * 2 + 1] =(acc_1)
                    C[ii * I_BLK + i * 4 + 1, jj * J_BLK + j * 2 + 0] =(acc_2)
                    C[ii * I_BLK + i * 4 + 1, jj * J_BLK + j * 2 + 1] = (acc_3)
                    C[ii * I_BLK + i * 4 + 2, jj * J_BLK + j * 2 + 0] =(acc_4)
                    C[ii * I_BLK + i * 4 + 2, jj * J_BLK + j * 2 + 1] = (acc_5)
                    C[ii * I_BLK + i * 4 + 3, jj * J_BLK + j * 2 + 0] = (acc_6)
                    C[ii * I_BLK + i * 4 + 3, jj * J_BLK + j * 2 + 1] = (acc_7)
        for i in range(I_BLK):
            for j in range((B.shape[0] // J_BLK) * J_BLK, B.shape[0]):
                acc_0 = init_val
                for k in range(A.shape[1]):
                    acc_0 += (A[ii * I_BLK + i, k] - B[j, k]) ** 2
                C[ii * I_BLK + i, j] = (acc_0)

    for i in range((A.shape[0] // I_BLK) * I_BLK, A.shape[0]):
        for j in range(B.shape[0]):
            acc_0 = init_val
            for k in range(A.shape[1]):
                acc_0 += (A[i, k] - B[j, k]) ** 2
            C[i, j] = (acc_0)

    return C



@numba.njit(fastmath=True)
def get_closest_for_inputs(X, W):
    min_dist_args = np.zeros(X.shape[0], dtype=np.int64)#
    min_dists = np.zeros(X.shape[0], dtype = np.int64)
    batch_len = 5000

    for b in range(len(X)//batch_len + 1):

        pdists = sq_eucl_opt(X[b * batch_len : (b+1) * batch_len], W)

        for i in range(pdists.shape[0]):

            min_dist_args[b*batch_len + i] = pdists[i].argmin()
            min_dists[b*batch_len + i] = pdists[i].min()
    return min_dist_args, min_dists



@numba.njit(fastmath=True)
def get_fast_pairwise(X, W, k):
    min_dist_args = np.zeros((X.shape[0], k), dtype=np.int64)  #
    min_dists = np.zeros((X.shape[0],k), dtype=np.float32)
    batch_len = 5000

    for b in range(len(X) // batch_len + 1):

        pdists = sq_eucl_opt(X[b * batch_len: (b + 1) * batch_len], W)

        for i in range(pdists.shape[0]):
            min_dist_args[b * batch_len + i] = get_closest(pdists[i],k)
            min_dists[b * batch_len + i] = pdists[i][min_dist_args[b * batch_len + i]]
    return min_dist_args, min_dists





# @numba.njit(fastmath=True)
def get_reverse_index(bmus, g_size):
    index = [[] for _ in range(g_size)]
    for i in range(len(bmus)):
        index[bmus[i]].append(i)
    return index

# @numba.jit(fastmath=True)
def combine_neighbors(index, neighbors):
    candidates = []

    for i in range(len(neighbors)):
        candidates.extend(index[neighbors[i]])
    return np.array(candidates)

def get_neighbor_lists(G):
    shp = np.arange(G.shape[0])

    neighbor_tree = []#[[] for _ in range(len(G))]

    for i in range(len(G)):
        bix = np.logical_or(G[i] == 1 , G.T[i] == 1)
        neighbor_tree.append(shp[bix])
    return neighbor_tree

# @numba.njit(fastmath=True)
def get_fast_knn(X_pc, W_pc, x_k, G):
    print('getting knns')
    closest_cvs, _ = get_closest_for_inputs(X_pc, W_pc)
    reverse_index = get_reverse_index(closest_cvs, G.shape[0])
    knns = np.zeros((X_pc.shape[0], x_k))
    knn_dists = np.zeros(knns.shape, dtype=np.float32)
    values = np.zeros(X_pc.shape[0] * x_k, dtype=np.float32)
    rows = np.zeros(X_pc.shape[0] * x_k, dtype=np.int32)
    cols = np.zeros(X_pc.shape[0] * x_k, dtype=np.int32)
    X_r = X_pc.reshape(X_pc.shape[0], 1, X_pc.shape[1])
    neighbor_tree = get_neighbor_lists(G)
    for i in range(X_pc.shape[0]):
        # G_neis = G[closest_cvs[i]]
        neighbors = neighbor_tree[closest_cvs[i]]
        candidates = combine_neighbors(reverse_index, neighbors)
        knn_entry, dists = get_fast_pairwise(X_r[i], X_pc[candidates], x_k)
        knns[i] = candidates[knn_entry[0]]
        knn_dists[i] = dists[0]
        start_ix = i * x_k
        end_ix = (i+1)  * x_k

        values[start_ix:end_ix] = knn_dists[i]/(knn_dists[i][-1]+0.0001)
        rows[start_ix:end_ix] = np.ones(x_k) * i
        cols[start_ix:end_ix] = knns[i]

    return values, rows, cols

@numba.njit('f4(f4,f4)', fastmath=True, )
def positive_clip(x, v):
    if x >= v:
        return v
    elif x <= -v:
        return -v
    else:
        return x


def find_spread_tightness(spread, min_dist):
    def curve(x, a, b):
        return 1. / (1. + a * x ** (2 * b))

    xv = np.linspace(0, spread * 3, 300)
    yv = np.zeros(xv.shape)
    yv[xv < min_dist] = 1.
    yv[xv >= min_dist] = np.exp(- (xv[xv >= min_dist] - min_dist) / spread)
    params, covar = curve_fit(curve, xv, yv)
    params = params.astype(np.float32)
    return params[0], params[1]


@numba.njit('f4(f4[:], f4[:])', fastmath=True, )
def rdist(x, y):
    dist = 0
    for i in range(len(x)):
        dist += pow(x[i] - y[i], 2)
    return dist


@numba.njit(fastmath = True)
def bulk_grow_with_drifters(shp, E_q, thresh_g, drifters, G, W, Y):
    growth_size = max(0, len(shp[E_q >= thresh_g]) - len(drifters))

    if growth_size:
        oldG = G
        oldW = W
        oldY = Y
        oldE = E_q
        old_size = oldW.shape[0]
        W = np.zeros((W.shape[0] + growth_size, W.shape[1]), dtype=np.float32)
        W[:-growth_size] = oldW

        Y = np.zeros((Y.shape[0] + growth_size, Y.shape[1]), dtype=np.float32)
        Y[: -growth_size] = oldY

        G = np.zeros((G.shape[0] + growth_size, G.shape[1] + growth_size), dtype=np.float32)
        G[:-growth_size][:, :-growth_size] = oldG

        E_q = np.zeros(E_q.shape[0] + growth_size, dtype=np.float32)
        E_q[:-growth_size] = oldE

        grown = 0
        growing_nodes = shp[oldE >= thresh_g]

        shp = np.arange(G.shape[0]).astype(np.int32)

        if len(growing_nodes) >= 1:
            for k in range(len(growing_nodes)):
                b = growing_nodes[k]
                ''' If no reusable nodes create new nodes'''
                closests = shp[G[b] == 1]
                if len(closests) == 0:
                    drifters = np.append(drifters, b)
                    continue
                h_bias = 1e-1
                l_bias = 1e-8
                W_n = (1-h_bias) * W[b] + h_bias * (W[closests].sum(axis=0) / len(closests))
                Y_n = (1-l_bias) * Y[b] + l_bias * (Y[closests].sum(axis=0) / len(closests))
                if grown >= len(drifters):
                    W[old_size + grown - len(drifters)] = W_n

                    Y[old_size + grown - len(drifters)] = Y_n
                    ''' connect neighbors to the new node '''
                    G[old_size + grown - len(drifters)][b] = 1
                    G[b][old_size + grown - len(drifters)] = 0
                    G[old_size + grown - len(drifters)][closests] = 1
                    G[closests][:, old_size + grown - len(drifters)] = 0

                    ''' Append new error. '''
                    E_q[old_size + grown - len(drifters)] = 0
                else:
                    '''replace unusable nodes with new nodes'''
                    W[drifters[grown]] = W_n
                    Y[drifters[grown]] = Y_n
                    G[drifters[grown]][b] = 1
                    G[b][drifters[grown]] = 0
                    G[drifters[grown]][closests] = 1
                    G[closests][:, drifters[grown]] = 0
                G[b][closests] = 0
                G[closests][:, b] = 0
                E_q[closests] *= 0.
                grown += 1

    return W, G, Y, E_q

# @numba.njit(fastmath = True)
def bulk_grow_with_drifters_duplex(shp, E_q , set_ix, thresh_g, drifters, G, W_0, W_1, Y):
    growth_size = max(0, len(shp[E_q[set_ix] >= thresh_g]) - len(drifters))

    if growth_size:
        oldG = G
        oldW_0 = W_0
        oldW_1 = W_1
        oldY = Y
        oldE = E_q
        old_size = oldW_0.shape[0]
        W_0 = np.zeros((W_0.shape[0] + growth_size, W_0.shape[1]), dtype=np.float32)
        W_0[:-growth_size] = oldW_0
        W_1 = np.zeros((W_1.shape[0] + growth_size, W_1.shape[1]), dtype=np.float32)
        W_1[:-growth_size] = oldW_1

        Y = np.zeros((Y.shape[0] + growth_size, Y.shape[1]), dtype=np.float32)
        Y[: -growth_size] = oldY

        G = np.zeros((G.shape[0] + growth_size, G.shape[1] + growth_size), dtype=np.float32)
        G[:-growth_size][:, :-growth_size] = oldG

        E_q_0 = np.zeros(E_q[0].shape[0] + growth_size, dtype=np.float32)
        E_q_0[:-growth_size] = oldE[0]
        E_q_1 = np.zeros(E_q[1].shape[0] + growth_size, dtype=np.float32)
        E_q_1[:-growth_size] = oldE[1]

        grown = 0
        growing_nodes = shp[oldE[set_ix] >= thresh_g]

        shp = np.arange(G.shape[0]).astype(np.int32)

        if len(growing_nodes) >= 1:
            for k in range(len(growing_nodes)):
                b = growing_nodes[k]
                ''' If no reusable nodes create new nodes'''
                closests = shp[G[b] == 1]
                if len(closests) == 0:
                    drifters = np.append(drifters, b)
                    continue
                h_bias = 1e-1
                l_bias = 1e-8
                W_n_0 = (1-h_bias) * W_0[b] + h_bias * (W_0[closests].sum(axis=0) / len(closests))
                W_n_1 = (1-h_bias) * W_1[b] + h_bias * (W_1[closests].sum(axis=0) / len(closests))

                Y_n = (1-l_bias) * Y[b] + l_bias * (Y[closests].sum(axis=0) / len(closests))
                if grown >= len(drifters):
                    W_0[old_size + grown - len(drifters)] = W_n_0
                    W_1[old_size + grown - len(drifters)] = W_n_1

                    Y[old_size + grown - len(drifters)] = Y_n
                    ''' connect neighbors to the new node '''
                    G[old_size + grown - len(drifters)][b] = 1
                    G[b][old_size + grown - len(drifters)] = 0
                    G[old_size + grown - len(drifters)][closests] = 1
                    G[closests][:, old_size + grown - len(drifters)] = 0

                    ''' Append new error. '''
                    E_q_0[old_size + grown - len(drifters)] = 0
                    E_q_1[old_size + grown - len(drifters)] = 0

                else:
                    '''replace unusable nodes with new nodes'''
                    W_0[drifters[grown]] = W_n_0
                    W_1[drifters[grown]] = W_n_1

                    Y[drifters[grown]] = Y_n
                    G[drifters[grown]][b] = 1
                    G[b][drifters[grown]] = 0
                    G[drifters[grown]][closests] = 1
                    G[closests][:, drifters[grown]] = 0
                G[b][closests] = 0
                G[closests][:, b] = 0
                E_q_0[closests] *= 0.
                E_q_1[closests] *= 0.

                grown += 1
            E_q = np.array([E_q_0, E_q_1])
    return W_0, W_1, G, Y, E_q

@numba.njit('f4(f4, f4)', fastmath=True)
def get_so_rate(tau, sigma):
    return  np.exp(-sigma * tau ** 2)

@numba.njit('i4[:](f4[:], i8)')
def get_closest(dists_2, k):
    neinds = [1 for _ in range(k)]
    n_W = len(dists_2)
    for i in range(n_W):
        dist2 = dists_2[i]
        ''' place dist2 in the neighbor list'''
        e = dist2
        if i < k or dists_2[neinds[-1]] > e:
            end = min(k - 1, i)
            mid = (end + 1) // 2
            beg = 0
            while end - beg and not (mid == beg or mid == end):
                if dists_2[neinds[mid]] > e:
                    end = mid
                else:
                    beg = mid
                mid = beg + (end - beg) // 2
            offset = dists_2[neinds[mid]] < e
            neinds.insert(mid + offset, i)
            neinds = neinds[:k]

    return np.array(neinds, dtype=np.int32)


@numba.njit(fastmath=True, )
def train_for_batch_batch(X_presented, pdist_matrix, i, max_its, lrst, lrdec, im_neix, W, max_epochs_per_sample, G, epsilon, min_strength,
                           shp, Y, ns_rate, alpha, beta, rng_state, E_q, lr_sigma, reduced_lr):


    taus = ((i * X_presented.shape[0] + np.arange(len(X_presented)).astype(np.float32)) * 1. / (max_its * X_presented.shape[0]))
    lrs = (1-taus)* reduced_lr
    so_lr = lrst * get_so_rate(i * 1. / max_its, lr_sigma)
    nei_len = np.int32(min(im_neix, W.shape[0]))

    for k in range(len(X_presented)):
        x = X_presented[k]
        dist_H = pdist_matrix[k]
        neilist = get_closest(dist_H, nei_len)

        b = neilist[0]
        G[b] *= epsilon

        G[b][neilist] = 1.
        G[b][G[b] < min_strength] = 0
        G[:, b][G[:, b] < min_strength] = 0
        nei_bin = (G[b] + G[:, b]) > 0
        neighbors = shp[nei_bin]

        denom = dist_H[neilist[-1]]

        epoch_vector = max_epochs_per_sample * ((G[b] + G[:, b]) / 2. + 1)
        neg_epoch_vector = ns_rate* (1 - (G[b] + G[:, b]) / 2.) + 1

        '''x, so_lr, b, W, hdist_nei, Y, alpha, beta , lr,  rng_state):'''
        W[neighbors], Y = train_neighborhood(x, so_lr, b, neighbors.astype(np.int32), W[neighbors],
                                             dist_H[neighbors] / denom,
                                             Y, ns_rate, alpha, beta,
                                             lrs[k], rng_state, epoch_vector.astype(np.int32),
                                             neg_epoch_vector.astype(np.int32))
        E_q[b] += dist_H[b]#/denom

    return W, Y, G, E_q


@numba.njit(
    'UniTuple(f4[:, :], 2)(f4[:], f4,  i4, i4[:],  f4[:,:], f4[:], f4[:,:], i4, f4, f4, f4, i8[:], i4[:], i4[:])',
    fastmath=True, )
def train_neighborhood(x, so_lr, b, neighbors, W, hdist_nei, Y, ns_rate, alpha, beta, lr, rng_state, epoch_vector,
                       neg_epoch_vector):
    hdists = hdist_nei
    y_b = Y[b]
    ''' Self Organizing '''
    sigma = 1
    for j in range(W.shape[0]):
        hdist = hdists[j]

        h_pull_grad =  1. if neighbors[j] == b else np.exp(-sigma * hdist)
        for i in range(x.shape[0]):
            W[j][i] += h_pull_grad * so_lr * (x[i] - W[j][i])

    '''negative embedding'''

    for j in range(W.shape[0]):

        epochs = epoch_vector[neighbors[j]]
        for e in range(epochs):
            for p in range(neg_epoch_vector[neighbors[j]]):
                n = fast_random_integer(rng_state) % Y.shape[0]
                ldist_sq = rdist(y_b, Y[n])
                push_grad = (2 * beta)
                denom = 1 + alpha * pow(ldist_sq, beta)
                push_grad /= denom
                for i in range(y_b.shape[0]):
                    if ldist_sq > 0.:
                        y_b[i] += positive_clip(push_grad / (ldist_sq + 0.001) * (y_b[i] - Y[n][i]),
                                                4) * lr
                        Y[n][i] -= positive_clip(push_grad / (ldist_sq + 0.001) * (y_b[i] - Y[n][i]),
                                                4) * lr
                    elif b == n:
                        continue
                    else:
                        y_b[i] += lr * 4
            Y_j = Y[neighbors[j]]
            ldist_sq = rdist(y_b, Y_j)
            '''deltas of the embedding are multiplied by a regularization term equal to the euclidean distance'''
            pull_grad = (2 * alpha * beta * pow(ldist_sq, beta - 1))
            denom = (1 + alpha * pow(ldist_sq, beta))
            pull_grad /= denom
            for i in range(y_b.shape[0]):
                Y_j[i] += positive_clip(pull_grad * (y_b[i] - Y_j[i]), 4) * lr
                y_b[i] -= positive_clip(pull_grad * (y_b[i] - Y_j[i]), 4) * lr

    return W, Y

@numba.njit('f4[:,:]( f4[:,:], f4[:, :], i4, i4, f4, f4, i8[:], f4 )', fastmath=True, )
def embed_batch_epochs(Y, G, max_its, i_st, alpha, beta, rng_state, agility):
    shp = np.arange(G.shape[0]).astype(np.int32)
    P_matrix = (G + G.T) / 2.
    tau = (i_st) * 1. / (max_its)
    tau_end = 1.
    epochs_per_sample = ((P_matrix)).astype(np.int32)
    dampening_coefficient = 1 if G.shape[0] < 50 else 1
    starting_lr = (1 - tau) * dampening_coefficient
    ending_lr = (1 - tau_end) * dampening_coefficient
    epochs_per_negative_sample = ((P_matrix.max() - P_matrix) * 5).astype(np.int32)
    totits = np.sum(epochs_per_sample)
    updated = 0

    pos_edges = np.where(epochs_per_sample)
    n_edges = len(pos_edges[0])
    for p in range(n_edges):
        j = pos_edges[0][p]
        lr = (starting_lr + (ending_lr - starting_lr) * (updated * 1. / totits)) * agility
        neighbors = shp[G[j] + G[:, j] > 0]
        for k in range(neighbors.shape[0]):
            epochs = epochs_per_sample[j][neighbors[k]]
            epochs_per_sample[j][neighbors[k]] = 0
            neg_epochs = int(epochs_per_negative_sample[j][neighbors[k]])
            for ep in range(epochs):
                for negs in range(neg_epochs):
                    n = fast_random_integer(rng_state) % Y.shape[0]
                    current_neg_epochs = epochs_per_negative_sample[j][n]
                    epochs_per_negative_sample[j][n] -= 1
                    if current_neg_epochs > 0:
                        ldist_sq = rdist(Y[j], Y[n])
                        push_grad = (2 * beta)
                        denom = 1 + alpha * pow(ldist_sq, beta)
                        push_grad /= denom
                        for i in range(Y[j].shape[0]):
                            if ldist_sq > 0.:
                                Y[j][i] += positive_clip(push_grad / (ldist_sq + 0.001) * (Y[j][i] - Y[n][i]), 4) * lr
                                Y[n][i] -= positive_clip(push_grad / (ldist_sq + 0.001) * (Y[j][i] - Y[n][i]), 4) * lr
                            elif j == n:
                                continue
                            else:
                                Y[j][i] += lr * 4
                ldist_sq = rdist(Y[j], Y[neighbors[k]])

                pull_grad = (2 * alpha * beta * pow(ldist_sq, beta - 1))
                denom = (1 + alpha * pow(ldist_sq, beta))
                pull_grad /= denom

                for i in range(Y[j].shape[0]):
                    Y[neighbors[k]][i] += positive_clip(pull_grad * (Y[j][i] - Y[neighbors[k]][i]), 4) * lr
                    Y[j][i] -= positive_clip(pull_grad * (Y[j][i] - Y[neighbors[k]][i]), 4) * lr
                updated += 1

    return Y



from distutils.core import setup
import setuptools

setup(
    name = "song_vis",
    version = "1.3.3.2",
    author = "Damith Senanayake",
    author_email = "dracusds123@gmail.com",
    description = "A dimensionality reduction method that retains a parametric model",
    packages = setuptools.find_packages(),
    download_url = 'https://github.com/damithsenanayake/SONG/archive/v1.1.1.tar.gz'
)